@@ -3,10 +3,29 @@ use strict;
 use warnings;
 use 5.010;
 
-our $VERSION = '0.26';
+our $VERSION = '0.27';
 
-our $PANDOC_VERSION;    # a string like '1.16'
-$PANDOC_VERSION ||= $ENV{PANDOC_VERSION};
+use Carp;
+use JSON qw(decode_json);
+use Scalar::Util qw(blessed reftype);
+use Pandoc::Walker qw(walk);
+use Pandoc::Version;
+
+our $PANDOC_VERSION; # a string like '1.16'
+$PANDOC_VERSION ||= eval { Pandoc::Version->new($ENV{PANDOC_VERSION}) };
+
+# internal variables
+my $PANDOC_BIN_MIN = Pandoc::Version->new('1.12.1');
+my $PANDOC_BIN_MAX = Pandoc::Version->new('1.18');
+
+my $PANDOC_API_MIN = Pandoc::Version->new('1.12.3');    # since pandoc 1.12.1
+my $PANDOC_API_DEFAULT = Pandoc::Version->new('1.17');  # since pandoc 1.18
+
+sub pandoc_version() {
+    return $PANDOC_BIN_MAX unless defined $PANDOC_VERSION;
+    (blessed $PANDOC_VERSION and $PANDOC_VERSION->isa('Pandoc::Version'))
+        ? $PANDOC_VERSION : Pandoc::Version->new($PANDOC_VERSION)
+}
 
 our %ELEMENTS = (
 
@@ -24,6 +43,7 @@ our %ELEMENTS = (
     Table          => [ Block => qw(caption alignment widths headers rows) ],
     Div            => [ Block => qw(attr content) ],
     Null           => ['Block'],
+    LineBlock      => [ Block => qw(content) ],
 
     # INLINE ELEMENTS
     Str         => [ Inline => 'content' ],
@@ -66,15 +86,10 @@ foreach (
     $ELEMENTS{$_} = ['Inline'];
 }
 
-use Carp;
-use JSON qw(decode_json);
-use Scalar::Util qw(blessed reftype);
-use Pandoc::Walker qw(walk);
-
 use parent 'Exporter';
 our @EXPORT = (
     keys %ELEMENTS,
-    qw(Document attributes metadata citation pandoc_json pandoc_query)
+    qw(Document attributes metadata citation pandoc_version pandoc_json pandoc_query)
 );
 our @EXPORT_OK = ( @EXPORT, 'element' );
 
@@ -123,12 +138,75 @@ sub element {
     &$name(@_);
 }
 
-sub Document($$) {
-    @_ == 2 or croak "Document expects 2 arguments, but given " . scalar @_;
-    my $meta = metadata(shift);
-    return bless [ { unMeta => $meta }, shift ], 'Pandoc::Document';
+sub Document {
+
+    my $arg = do {
+        if ( @_ == 1 ) {
+            my $reftype = reftype $_[0] // '';
+            if ( $reftype eq 'ARRAY') {
+                # old JSON format
+                {
+                    meta => $_[0]->[0]->{unMeta}, 
+                    blocks => $_[0]->[1], 
+                    api_version => 1.16,
+                }
+            } elsif ( $reftype eq 'HASH' ) {
+                $_[0]
+            } else {
+                croak 'Document: expect array or hash reference'
+            }
+        } elsif ( @_ == 2 ) {
+            # \%meta, \@blocks
+            { meta => $_[0], blocks => $_[1] }
+        } elsif ( @_ % 2 ) {
+            # odd number of args
+            croak "Document: too many or ambiguous arguments";
+        } else {
+            # even number of args: api_version as named parameter
+            { meta => shift, blocks => shift, @_ }
+        }
+    };
+    
+    # prefer haskell-style key but accept perl-style and abbreviated key
+    my $api_version = $arg->{'pandoc-api-version'} // $arg->{pandoc_api_version}
+      // $arg->{api_version} // $PANDOC_API_DEFAULT;
+    $api_version = Pandoc::Version->new( $api_version );
+
+    croak 'api_version must be >= 1.12.3' if $api_version < $PANDOC_API_MIN;
+
+    # We copy values here because $arg may not be a pure AST representation
+    my $doc = bless {
+        meta   => metadata( $arg->{meta} // {} ),
+        blocks => ( $arg->{blocks}       // [] ),
+        'pandoc-api-version' => $api_version,
+      },
+      'Pandoc::Document';
+
+    walk $doc, \&_bless_pandoc_element;
+
+    return $doc;
+
 }
 
+# internal helper method
+sub _bless_pandoc_element {
+    my $e = shift;
+    return $e unless ref $e;
+    return $e if blessed $e and $e->isa('Pandoc::Document::Element');
+
+    # TODO: run recursively via set_content (don't require 'walk')
+    if ( 'MetaMap' eq $e->{t} ) {
+        for my $v ( values %{ $e->{c} } ) {
+            _bless_pandoc_element( $v );
+        }
+    }
+
+    bless $e, 'Pandoc::Document::' . $e->{t};
+    $e->upgrade($e) if $e->can('upgrade');
+    return $e;
+}
+
+
 # specific accessors
 
 sub Pandoc::Document::DefinitionPair::term        { $_[0]->[0] }
@@ -160,20 +238,14 @@ sub citation($) {
 }
 
 sub metadata($) {
-    $_[0]; # TODO: issue #10 and #34
+    my $meta = shift; # TODO: issue #10 and #34
+    for my $v ( values %$meta ) {
+        $v = _bless_pandoc_element( $v );
+    }
+    $meta;
 }
 
 sub pandoc_json($) {
-    state $ast_to_element = sub {   # compile once, use repeatedly, keep private
-        my $class = 'Pandoc::Document::' . $_[0]->{t};
-        if ( 'MetaMap' eq $_[0]->{t} ) {
-            for my $v ( values %{ $_[0]->{c} } ) {
-                $_[1]->( $v, $_[1] );
-            }
-        }
-        return $class->new_from_ast( $_[0] );
-    };
-
     shift if $_[0] =~ /^Pandoc::/;
 
     my $ast = eval { decode_json( $_[0] ) };
@@ -182,24 +254,7 @@ sub pandoc_json($) {
         chomp $@;
         croak $@;
     }
-    return unless reftype $ast;
-
-    if ( reftype $ast eq 'ARRAY' ) {
-        my $meta = $ast->[0]->{unMeta};
-        for my $v ( values %$meta ) {
-            $v = $ast_to_element->( $v, $ast_to_element );
-        }
-        $ast = Document( $meta, $ast->[1] );
-    }
-    elsif ( reftype $ast eq 'HASH' and $ast->{t} ) {
-
-        # $ast = element( $ast->{t}, $ast->{c} );
-        $ast = $ast_to_element->( $ast, $ast_to_element );
-    }
-
-    walk $ast, $ast_to_element;
-
-    return $ast;
+    return Document $ast;
 }
 
 *pandoc_query = *Pandoc::Walker::query;
@@ -212,15 +267,17 @@ sub pandoc_json($) {
     use strict;
     our $VERSION = '0.04';
     our @ISA = ('Pandoc::Document::Element');
+    sub blocks;
     sub name { 'Document' }
     sub meta {
-        $_[0]->[0]->{unMeta} = Pandoc::Elements::metadata($_[1]) if @_ > 1;
-        $_[0]->[0]->{unMeta}
+        $_[0]->{meta} = Pandoc::Elements::metadata($_[1]) if @_ > 1;
+        $_[0]->{meta};
     }
-    sub content { 
-        $_[0]->[1] = $_[1] if @_ > 1; 
-        $_[0]->[1] 
+    sub content {
+        $_[0]->{blocks} = $_[1] if @_ > 1;
+        $_[0]->{blocks};
     }
+    *blocks = \&content;
     sub is_document { 1 }
     sub metavalue {
         my $meta = $_[0]->meta;
@@ -229,6 +286,13 @@ sub pandoc_json($) {
     sub string {
         join '', map { $_->string } @{$_[0]->content}
     }
+    sub api_version {
+        my $self = shift;
+        if ( @_ ) {
+            $self->{'pandoc-api-version'} = Pandoc::Version->new( $_[0] )
+        }
+        return $self->{'pandoc-api-version'};
+    }
 }
 
 {
@@ -295,8 +359,6 @@ sub pandoc_json($) {
     *query     = *Pandoc::Walker::query;
     *transform = *Pandoc::Walker::transform;
 
-    sub new_from_ast { bless $_[1] => $_[0] }
-
     sub string {
 
         # TODO: fix issue #4 to avoid this duplication
@@ -471,23 +533,42 @@ sub pandoc_json($) {
     sub url   { $_[0]->{c}->[-1][0] }
     sub title { $_[0]->{c}->[-1][1] }
 
-    sub new_from_ast {
-        my ( $class, $ast ) = @_;
-        if ( 2 == @{ $ast->{c} } ) {
-
-            # prepend attributes to old-style ast
-            unshift @{ $ast->{c} }, [ "", [], [] ];
-        }
-        return bless $ast => $class;
+    sub upgrade {
+        # prepend attributes to old-style ast
+        unshift @{ $_[0]->{c} }, [ "", [], [] ]
+            if 2 == @{ $_[0]->{c} };
     }
 }
 
 # Special TO_JSON methods to coerce data to int/number/Boolean as appropriate
-# and to downgrade document model for Pandoc < 1.16
+# and to downgrade document model depending on pandoc_version
+
+sub Pandoc::Document::to_json {
+    my ($self) = @_;
+
+    local $Pandoc::Elements::PANDOC_VERSION =
+        $Pandoc::Elements::PANDOC_VERSION // do {
+        if ( $self->api_version < 1.17 ) {
+            $self->api_version < 1.16 ? '1.12.1' : '1.16'
+        } else {
+            '1.18'
+        }
+    };
+    Pandoc::Document::Element::to_json($self);
+}
+
+sub Pandoc::Document::TO_JSON {
+    my ( $self ) = @_;
+    return Pandoc::Document::Element::TO_JSON(
+        $self->api_version ge '1.17'
+        ? $self
+        : [ { unMeta => $self->{meta} }, $self->{blocks} ]
+    );
+}
 
 sub Pandoc::Document::SoftBreak::TO_JSON {
-    if ( $Pandoc::Elements::PANDOC_VERSION
-        and ( $Pandoc::Elements::PANDOC_VERSION lt '1.16' ) ) {
+    #say STDERR "pandoc_version ".pandoc_version;
+    if ( pandoc_version < '1.16' ) {
         return { t => 'Space', c => [] };
     } else {
         return { t => 'SoftBreak', c => [] };
@@ -496,10 +577,8 @@ sub Pandoc::Document::SoftBreak::TO_JSON {
 
 sub Pandoc::Document::LinkageRole::TO_JSON {
     my $ast = Pandoc::Document::Element::TO_JSON( $_[0] );
-    if ( $Pandoc::Elements::PANDOC_VERSION
-        and ( $Pandoc::Elements::PANDOC_VERSION lt '1.16' ) )
-    {
-        # remove attributes from new-style ast
+    if ( pandoc_version < 1.16 ) {
+        # remove attributes
         $ast->{c} = [ @{ $ast->{c} }[ 1, 2 ] ];
     }
     return $ast;
@@ -579,6 +658,25 @@ sub Pandoc::Document::Cite::TO_JSON {
     return $ast;
 }
 
+sub Pandoc::Document::LineBlock::TO_JSON {
+    my $ast     = Pandoc::Document::Element::TO_JSON( $_[0] );
+    my $content = $ast->{c};
+
+    for my $line ( @$content ) {
+
+        # Convert spaces at the beginning of each line
+        # to Unicode non-breaking spaces, because pandoc does.
+        next unless $line->[0]->{t} eq 'Str';
+        $line->[0]->{c} =~ s{^(\x{20}+)}{ "\x{a0}" x length($1) }e;
+    }
+
+    return $ast if pandoc_version >= 1.18;
+
+    my $c = [ map { ; @$_, LineBreak() } @{$content} ];
+    pop @$c;    # remove trailing line break
+    return Para( $c )->TO_JSON;
+}
+
 1;
 __END__
 
@@ -605,14 +703,18 @@ The output of this script C<hello.pl>
     use Pandoc::Elements;
     use JSON;
 
-    print Document({
+    print Document(
+        {
             title => MetaInlines [ Str "Greeting" ]
-        }, [
+        },
+        [
             Header( 1, attributes { id => 'top' }, [ Str 'Hello' ] ),
             Para [ Str 'Hello, world!' ],
-        ])->to_json;
+        ],
+        api_version => '1.17.0.4'
+    )->to_json;
 
-can be converted for instance to HTML with via
+can be converted for instance to HTML via
 
     ./hello.pl | pandoc -f json -t html5 --standalone
 
@@ -653,8 +755,8 @@ as types in other document elements.
 
 =item
 
-The helper following functions C<pandoc_json>, C<attributes>, C<citation>, and
-C<element>.
+The following helper functions C<pandoc_json>, C<pandoc_version>,
+C<attributes>, C<citation>, and C<element>.
 
 =back
 
@@ -692,6 +794,21 @@ construct such hash by filling in default values and using shorter field names
 
     [see @foo p. 42]
 
+=head2 pandoc_version
+
+Return expected version number of pandoc executable to be used for serializing
+documents with L<to_json|/to_json>. The abstract syntax tree of Pandoc
+documents, reflected in this module, changes slightly between some releases of
+pandoc (for instance pandoc 1.16 introduced attributes to L<Link|/Link> and
+L<Image|/Image> elements).  Package variable C<$PANDOC_VERSION> can be used to
+set the expected version. By default it is set from an environment variable of
+same name. This method returns the current value of the variable or the most
+recent version reliably supported by this module as instance of
+L<Pandoc::Version>.
+
+See also method C<version> of module L<Pandoc> to get the current version of
+pandoc executable on your system.
+
 =head2 element( $name => $content )
 
 Create a Pandoc document element of arbitrary name. This function is only
@@ -700,10 +817,11 @@ exported on request.
 =head1 ELEMENTS AND METHODS
 
 Document elements are encoded as Perl data structures equivalent to the JSON
-structure, emitted with pandoc output format C<json>. All elements are blessed
-objects that provide L<common element methods|/COMMON METHODS> (all elements),
-L<attribute methods|/ATTRIBUTE METHODS> (elements with attributes), and
-additional element-specific methods.
+structure, emitted with pandoc output format C<json>. This JSON structure is
+subject to minor changes between L<versions of pandoc|/pandoc_version>.  All
+elements are blessed objects that provide L<common element methods|/COMMON
+METHODS> (all elements), L<attribute methods|/ATTRIBUTE METHODS> (elements with
+attributes), and additional element-specific methods.
 
 =head2 COMMON METHODS
 
@@ -714,10 +832,8 @@ Return the element as JSON encoded string. The following are equivalent:
     $element->to_json;
     JSON->new->utf8->canonical->convert_blessed->encode($element);
 
-Note that the JSON format changed from Pandoc 1.15 to Pandoc 1.16 by introduction
-of attributes to L<Link|/Link> and L<Image|/Image> elements. Since Pandoc::Elements
-0.16 the new format is serialized by default. Set the package variable or
-environment variable C<PANDOC_VERSION> to 1.15 or below to use the old format.
+Note that the suitable JSON format depends on the pandoc executable version.
+See L</PANDOC VERSION> for details.
 
 =head3 name
 
@@ -809,17 +925,94 @@ append identifier or class, respectively.
 
 =head3 Document
 
-Root element, consisting of metadata hash (C<meta>) and document element array
-(C<content>).
+Root element, consisting of metadata hash (C<meta>), document element array
+(C<content>=C<blocks>) and optional C<api_version>. The constructor accepts
+either two arguments and an optional named parameter C<api_version>:
 
-    Document $meta, [ @blocks ]
+    Document { %meta }, [ @blocks ], api_version => $version_string
 
-Document C<metavalue> returns a copy of the metadata hash with all L<metadata
-elements|/METADATA ELEMENTS> flattened to unblessed values:
+or a hash with three fields for metadata, document content, and an optional
+pandoc API version:
+
+    {
+        meta               => { %metadata },
+        blocks             => [ @content ],
+        pandoc-api-version => [ $major, $minor, $revision ]
+    }
+
+The latter form is used as pandoc JSON format since pandoc release 1.18. If no
+api version is given, it will be set 1.17 which was also introduced with pandoc
+release 1.18.
+
+A third ("old") form is accepted for compatibility with pandoc JSON format
+before release 1.18 and since release 1.12.1: an array with two elements for
+metadata and document content respectively.
+
+    [ { unMeta => { %meta } }, [ @blocks ] ]
+
+The api version is set to 1.16 in this case, but older versions down to 1.12.3
+used the same format.
+
+Document elements provide the following special methods in addition to
+L<common element methods|/COMMON METHODS>:
+
+=over
+
+=item C<api_version>
+
+returns a L<Pandoc::Version|Pandoc::Version> object, or takes a string like
+C<'1.17.0.4'> to set the value. Note that the actual number of fields in the
+version number may be greater than three.
+
+Beginning with version 1.18 pandoc will not decode a JSON AST
+representation unless the major and minor version numbers
+stored in the C<pandoc-api-version> field match those
+built into that version of pandoc.
+
+To determine the API version required by the version of the pandoc
+executable you are running run pandoc with the C<--version> option
+and check which version of the C<pandoc-types> library pandoc was
+compiled with. As of pandoc 1.18 this is the same as the API version
+number required in the JSON AST representation.
+
+If the API version number of the Document object is less than
+C<1.17.0.4>, the API version required by pandoc 1.18, the
+Document C<to_json> method will emit the old-style (pre-pandoc-
+1.18) array-based AST representation. When writing filters you
+should normally just rely on the API version value obtained from
+pandoc, if any, since pandoc expects to receive the same JSON
+format as it emits.
+
+If no API version number is present in the arguments given to the
+Document constructor the value of the object returned by the
+C<api_version> method will default to the dummy value C<0> (zero).
+Thus the object returned by the C<api_version> method is
+always safe to compare with another Pandoc::Version object or a string
+with a version number. When checking whether pandoc expects the new
+or the old AST representation it is however safer to check with
+C<< $document->api_version ge '1.17.0.4' >>. Since earlier versions
+of Pandoc::Elements do not support the C<< $document->api_version >>
+method you should wrap such a check in an C<eval> block if
+your program should be able to run under earlier versions.
+
+=item C<content> or C<blocks>
+
+Get or set the array of L<block elements|/BLOCK ELEMENTS> of the
+document.
+
+=item C<meta>
+
+Return document L<metadata elements|/METADATA ELEMENTS>.
+
+=item C<metavalue>
+
+Returns a copy of the metadata hash with all L<metadata elements|/METADATA
+ELEMENTS> flattened to unblessed values:
 
     $doc->metavalue   # equivalent to
     { map { $_ => $doc->meta->{$_}->metavalue } keys %{$doc->meta} }
 
+=back
 
 =head2 BLOCK ELEMENTS
 
@@ -875,6 +1068,17 @@ Horizontal rule
 
     HorizontalRule
 
+=head3 LineBlock
+
+List of lines (C<content>), each a list of L<inlines|/INLINE ELEMENTS>.
+
+    LineBlock [ @lines ]
+
+This element was added in pandoc 1.18. Before it was represented L<Para|/Para>
+elements with embedded L<LineBreak|/LineBreak> elements. This old serialization
+form can be enabled by setting C<$PANDOC_VERSION> package variable to a lower
+version number.
+
 =head3 Null
 
 Nothing
@@ -1023,11 +1227,11 @@ Soft line break
 
     SoftBreak
 
-Note that the C<SoftBreak> element was added in Pandoc 1.16 to as a matter of
-editing convenience to preserve line breaks (as opposed to paragraph breaks)
-from input source to output. If you are going to feed a document containing
-C<SoftBreak> elements to Pandoc E<lt> 1.16 you will have to set the package
-variable or environment variable C<PANDOC_VERSION> to 1.15 or below.
+This element was added in pandoc 1.16 as a matter of editing convenience to
+preserve line breaks (as opposed to paragraph breaks) from input source to
+output. If you are going to feed a document containing C<SoftBreak> elements to
+Pandoc E<lt> 1.16 you will have to set the package variable or environment
+variable C<PANDOC_VERSION> to 1.15 or below.
 
 =head3 Space
 
@@ -1164,7 +1368,9 @@ C<DefaultDelim>, C<Period>, C<OneParen>, C<TwoParens>
 
 =head1 SEE ALSO
 
-L<Pandoc> implements a wrapper around the pandoc executable.
+Perl module L<Pandoc> implements a wrapper around the pandoc executable.
+
+Similar libraries in other programming languages are listed at L<https://github.com/jgm/pandoc/wiki/Pandoc-wrappers-and-interfaces>.
 
 L<Text.Pandoc.Definition|https://hackage.haskell.org/package/pandoc-types/docs/Text-Pandoc-Definition.html>
 contains the original definition of Pandoc document data structure in Haskell.
